const Node = require('../parser/Node');

let nodeCount = 0;

/**
 * Create eagel recognized template JSON.
 * @param node {Node}
 * @param node.tag {String} Tag name.
 * @param node.attrs {Object}  Attributes.
 * @param node.children? {Array} Child nodes.
 * @param options? {Object} Options.
 */
module.exports = function generate(node, options = {}) {
  nodeCount = 0; // Reset node count
  return traverse(node, (n) => transformNode(n, options));
}

/**
 * Traverse in DFS.
 * @param node
 * @param fn
 */
function traverse(node, fn) {
  if (Array.isArray(node)) {
    return node.map((n) => traverse(n, fn)).filter(n => n);
  }

  const newNode = fn(node);
  if (node && Array.isArray(node.children)) {
    newNode.childNodes = traverse(node.children, fn);
  }
  return newNode;
}

/**
 * Transform Node to eagel json.
 * @param node
 * @param options
 * @return {Node}
 */
function transformNode(node, options) {
  const { componentAlias = {}, modules } = options;
  if (Array.isArray(modules)) {
    for (let i = 0, l = modules.length; i < l; i++) {
      if (typeof modules[i].transformNode === 'function')
        node = modules[i].transformNode(node, options);
    }
  }
  if (node) {
    if (typeof node === 'string') {
      // Wrap textNode to text element.
      if (!node.trim()) return null;

      nodeCount++;
      const tagName = 'text';
      let ret = {
        tagName,
        nodeId: '' + nodeCount,
        attributes: {
          value: [node],
        },
      };

      return ret;
    } else {
      nodeCount++;

      const tagName = node.tag;
      let ret = {
        tagName,
        nodeId: '' + nodeCount,
      };

      // Capticalized tag name means component.
      if (/^[A-Z]/.test(tagName)) {
        ret.isComponent = true;
        ret.componentAlias = {};
        const alias = componentAlias[tagName];
        if (alias) ret.componentAlias[tagName] = alias;
      }

      for (let key in node.attrs) {
        if (node.attrs.hasOwnProperty(key)) {
          const value = node.attrs[key];
          switch (key) {
            case 'ref':
            case 'style':
              ret[key] = value;
              break;

            case 'class':
            case 'className': {
              ret.classList = ret.classList || [];
              ret.classList = ret.classList.concat(value.split(' '));
              break;
            }

            default: {
              ret.attributes = ret.attributes || {};
              if (isBinding(value)) {
                ret.attributes[key] = { '@binding': getBinding(value) }
              } else {
                ret.attributes[key] = value;
              }
            }
          }
        }
      }

      if (node.tag === 'text') {
        const value = [];
        if (Array.isArray(node.children)) {
          for (let i = 0, l = node.children.length; i < l; i++) {
            if (typeof node.children[i] === 'string') {
              const v = node.children[i].trim();
              if (isBinding(v)) {
                value.push({ '@binding': getBinding(v) });
              } else {
                value.push(v);
              }
            }
          }

          // Remove children in case of textNode be processed.
          delete node.children;
        }
        ret.attributes = ret.attributes || {};
        ret.attributes.value = value;
      }

      return ret;
    }

  } else {
    return node;
  }
}

const BINDING_REG = /{{([^}]+)}}/;
function isBinding(exp) {
  return BINDING_REG.test(exp);
}
function getBinding(exp) {
  const ret = BINDING_REG.exec(exp);
  return ret ? ret[1] : '';
}
