const { resolve, extname, relative } = require('path');
const { existsSync, readFileSync } = require('fs-extra');
const { default: traverse, NodePath } = require('@babel/traverse');
const t = require('@babel/types');
const isClassComponent = require('./utils/isClassComponent');
const isFunctionComponent = require('./utils/isFunctionComponent');
const codegen = require('./codegen');
const transformStyle = require('./codegen/transformStyle');
const generateCodeByExpression = require('./codegen/generateCodeByExpression')
const { parse } = require('./parser');
const parseRender = require('./parser/parseRender');
const { parseClassState, parseUseState } = require('./parser/parseState');
const { parseElement, parserAdapter } = require('./parser/parseElement');
const modules = require('./modules');
const {
  RAX_COMPONENT,
  CREATE_COMPONENT,
  SAFE_CREATE_COMPONENT,
  EXPORTED_CLASS_DEF,
  HELPER_RUNTIME,
  RAX_PACKAGE
} = require('./constant');
const cwd = process.cwd();

const templateIdMap = {};
function getTemplateId(path) {
  if (templateIdMap.hasOwnProperty(path)) {
    return templateIdMap[path];
  } else {
    return templateIdMap[path] = String(Object.keys(templateIdMap).length + 1);
  }
}
/**
 * JSX Compiler
 * transform JSX DSL to Eagel.
 * ref. https://yuque.antfin-inc.com/weexcore/weex-eagle/si936d
 */
module.exports = function transform(code, options = {}) {
  const { filePath, rootContext } = options;
  const ast = parse(code);
  const style = getStyle(ast, { filePath });
  const { node, components, componentAlias } = getTemplate(ast, { filePath });
  const data = getData(ast);

  return {
    body: codegen(node, { style, componentAlias, modules }),
    data,
    components,
    styles: transformStyle(style),
    // 需确认是否需要
    props: {},
  };
}

/**
 * Get axml template from babel AST.
 * @param ast {ASTElement} Babel AST structure.
 * @param options {Object} Get template options..
 */
function getTemplate(ast, options) {
  const { filePath } = options || {};
  const componentAlias = {};
  let node = null;
  let components = [];
  traverse(ast, {

    // Collect imported components.
    ImportDeclaration(path) {
      const { node } = path;
      const { source, specifiers } = node;
      const defaultSpecifier = specifiers.filter(n => t.isImportDefaultSpecifier(n))[0];

      let isValidJSXModule = false;
      let currentFilePath = filePath;
      currentFilePath = resolve(currentFilePath, '..', source.value);

      const ext = extname(currentFilePath);
      if (ext === '.js' || ext === '.jsx' && existsSync(currentFilePath)) {
        isValidJSXModule = true;
      } else if (existsSync(currentFilePath + '.js')) {
        currentFilePath = currentFilePath + '.js';
        isValidJSXModule = true;
      } else if (existsSync(currentFilePath + '.jsx')) {
        currentFilePath = currentFilePath + '.jsx';
        isValidJSXModule = true;
      }

      const code = isValidJSXModule
        ? readFileSync(currentFilePath, 'utf-8')
        : null;
      if (defaultSpecifier && code) {
        const filePath = currentFilePath;
        componentAlias[defaultSpecifier.local.name] = relative(cwd, filePath);
        const ast = parse(code);
        const style = getStyle(ast, { filePath });
        const { node, components: subComponents } = getTemplate(ast, { filePath });
        const data = getData(ast);

        if (Array.isArray(subComponents) && subComponents.length > 0) {
          components = components.concat(subComponents);
        }

        components.push({
          template: codegen(node, { style, componentAlias, modules }),
          templateId: getTemplateId(filePath),
          data,
          styles: transformStyle(style),
          name: relative(cwd, filePath),
          props: {},
        });
      }
    },

    // Detect exported component.
    ExportDefaultDeclaration(path) {
      const declarationPath = path.get('declaration');
      if (isClassComponent(declarationPath)) {
        const renderFnPath = getRenderMethod(declarationPath);
        if (renderFnPath) {
          // Rule restrict: allow one return statement.
          const [, error] = renderFnPath.node.body.body.filter(s => t.isReturnStatement(s));
          if (error) {
            // TODO: 报错需要带文件和行列信息
            throw Error('Only one return is allow in render method.');
          }

          node = parseRender(renderFnPath);

          // Remove render method at last.
          renderFnPath.remove();
        }
      } else if (isFunctionComponent(declarationPath)) {
        const returnStatementPath = getReturnStatement(declarationPath);
        node = parseElement(returnStatementPath.get('argument').node);

        // Remove return statement at last.
        returnStatementPath.remove();
      } else {
        throw Error('Can not find default export in file.');
      }
    },
  });

  return { node, components, componentAlias };
}

/**
 * Get the render function in class component.
 * @param path {NodePath} A nodePath that contains a render function.
 * @return {NodePath} Path to render function.
 */
function getRenderMethod(path) {
  const { node, scope } = path;
  if (t.isIdentifier(node)) {
    const binding = scope.getBinding(node.name);
    if (t.isVariableDeclarator(binding.path.node)) {
      return getRenderMethod(binding.path.get('init'));
    } else {
      return getRenderMethod(binding.path);
    }
  } else {
    let renderMethod = null;
    path.traverse({
      /**
       * Example:
       *   class {
       *     render() {}
       *   }
       */
      ClassMethod(classMethodPath) {
        const { node } = classMethodPath;

        if (t.isIdentifier(node.key, { name: 'render' })) {
          renderMethod = classMethodPath;
        }
      },
      /**
       * Example:
       *   class {
       *     render = function() {}
       *     render = () => {}
       *   }
       */
      ClassProperty(path) {
        // TODO: Unsupported.
      },
    });

    return renderMethod;
  }
}

/**
 * Get the render return statement in function component..
 * @param path {NodePath} A nodePath that contains a render return.
 * @return {NodePath} Path to return statement.
 */
function getReturnStatement(path) {
  const { node, scope } = path;
  if (t.isIdentifier(node)) {
    const binding = scope.getBinding(node.name);
    if (t.isVariableDeclarator(binding.path.node)) {
      return getReturnStatement(binding.path.get('init'));
    } else {
      return getReturnStatement(binding.path);
    }
  } else {
    let returnStatementPath = null;
    path.traverse({
      ReturnStatement: {
        exit(path) {
          returnStatementPath = path;
        }
      }
    });
    return returnStatementPath;
  }
}

/**
 * Get miniapp component js code from babel AST.
 * @param ast {ASTElement} Babel AST structure.
 * @return {String} Miniapp component js code.
 */
function getComponentJSCode(ast) {
  let delayToRemovePath;
  traverse(ast, {
    /**
     * 1. Add import declaration of helper lib.
     * 2. Rename scope's Component to other id.
     * 3. Add Component call expression.
     * 4. Transform 'rax' to 'rax/dist/rax.min.js' in case of 小程序开发者工具 not support `process`.
     */
    Program(path) {
      const importedIdentifier = t.identifier(CREATE_COMPONENT);
      const localIdentifier = t.identifier(SAFE_CREATE_COMPONENT);

      // import { createComponent as __create_component__ } from "jsx2mp-runtime";
      path.node.body.unshift(
        t.importDeclaration(
          [t.importSpecifier(localIdentifier, importedIdentifier)],
          t.stringLiteral(HELPER_RUNTIME)
        )
      );

      // Component(__create_component__(__class_def__));
      path.node.body.push(
        t.expressionStatement(
          t.callExpression(
            t.identifier(RAX_COMPONENT),
            [
              t.callExpression(
                t.identifier(SAFE_CREATE_COMPONENT),
                [t.identifier(EXPORTED_CLASS_DEF)]
              )
            ],
          )
        )
      );
    },

    ImportDeclaration(path) {
      if (t.isStringLiteral(path.node.source, { value: RAX_PACKAGE })) {
        delayToRemovePath = path;
      }
    },

    ExportDefaultDeclaration(path) {
      const declarationPath = path.get('declaration');
      if (isClassComponent(declarationPath)) {
        const { id, body, decorators } = declarationPath.node;
        path.replaceWith(
          t.variableDeclaration('var', [
            t.variableDeclarator(
              t.identifier(EXPORTED_CLASS_DEF),
              t.classExpression(id, null, body, decorators)
            )
          ])
        );
      }
    },
  });

  if (delayToRemovePath) delayToRemovePath.remove();

  return generateCodeByExpression(ast);
}

/**
 * Get style content from import declaration.
 * @param ast
 * @param filePath
 * @return {string}
 */
function getStyle(ast, { filePath }) {
  let ret = '';
  traverse(ast, {
    ImportDeclaration(path) {
      const { node } = path;
      const { source } = node;
      let currentFilePath = filePath;
      currentFilePath = resolve(currentFilePath, '..', source.value);
      if (extname(currentFilePath) === '.css') {
        ret += readFileSync(currentFilePath, 'utf-8');
      }
    },
  });
  return ret;
}

function getData(ast) {
  let ret = null;
  traverse(ast, {
    ExportDefaultDeclaration(path) {
      const declarationPath = path.get('declaration');
      if (isClassComponent(declarationPath)) {
        ret = parseClassState(declarationPath);
      } else if (isFunctionComponent(declarationPath)) {
        ret = parseUseState(declarationPath);
      }
    },
  });
  return ret;
}
