const t = require('@babel/types');
const convertASTToVariable = require('../utils/convertASTToVariable');

/**
 * Parse data from class state declaration.
 * eg:
 * export default class extends Rax.Component {
 *   state = { color: 'red' };
 * }
 * @param nodePath
 */
exports.parseClassState = function parseClassState(nodePath) {
  const state = {};
  nodePath.traverse({
    ClassProperty(path) {
      const { node } = path;
      if (t.isIdentifier(node.key, { name: 'state' })) {
        Object.assign(state, convertASTToVariable(node.value));
      }
    }
  });

  return state;
};

/**
 * Parse data from useState.
 * eg:
 * export default function() {
 *   const [ color, setColor ] = useState('red');
 * }
 * @param nodePath
 */
exports.parseUseState = function parseUseState(nodePath) {
  const state = {};
  nodePath.traverse({
    VariableDeclarator(path) {
      const { node, scope } = path;
      if (isUseStateCallExpression(node.init)) {
        let key, value;
        if (t.isArrayPattern(node.id)) {
          const valueId = node.id.elements[0];
          key = valueId.name;
        } else {
          throw new Error('useState should use like: const [ foo, setFoo ] = useState(defaultFoo);.');
        }

        const valueId = node.init.arguments[0];
        if (t.isIdentifier(valueId)) {
          value = getBindingValue(scope, valueId.name);
        } else if (t.isLiteral(valueId)) {
          value = valueId.value;
        } else {
          throw new Error('useState should pass literial value.');
        }

        state[key] = value;
      }
    }
  });

  return state;
};

function isUseStateCallExpression(node) {
  return t.isCallExpression(node) && t.isIdentifier(node.callee, { name: 'useState' })
}

function getBindingValue(scope, name) {
  const binding = scope.getBinding(name);
  if (binding && binding.path) {
    if (t.isLiteral(binding.path.node.init)) {
      return binding.path.node.init.value;
    } else if (t.isIdentifier(binding.path.node.init)) {
      return getBindingValue(scope, binding.path.node.init.name);
    }
  }
  throw new Error('Can not get binding value "' + name + '"');
}
